 
from psycopg import connect
from os import getenv
from datetime import datetime, timezone
from klp_commons.log.structuredMessage import StructuredMessage
from nanoid import generate
from logging import getLogger

fmt = StructuredMessage
message = 'postgresDBController'

class ControllerPostgresDB:

    def __init__(self):

        self.conn = None
        self.logger = getLogger('categorization.ControllerPosgrestDB')

        # Init ()
        self.logger.info('ControllerPosgrestDB.init()')
        self.logger.info('creating an instance of ControllerPosgrestDB')
        self.PG_HOST = getenv('PG_HOST')
        self.PG_DATABASE = getenv('PG_DATABASE')
        self.PG_USER = getenv('PG_USER')
        self.PG_PASSWORD = getenv('PG_PASSWORD')
        self.PG_PORT = getenv('PG_PORT')
        self.format_exter = '%Y-%m-%dT%H:%M:%S' # .%f%z'
        self.format_inter = '%Y-%m-%dT%H:%M:%S'
        self.logger.info('donecreating an instance of ControllerPosgrestDB')

    def get_conection(self) -> None:
        self.logger.info('call get_conection method ')

        try:
            self.conn = connect(
                host=self.PG_HOST,
                dbname=self.PG_DATABASE,
                user=self.PG_USER,
                password=self.PG_PASSWORD,
                port=self.PG_PORT,
            )

        except BaseException as e:
            self.conn.rollback()
            self.logger.exception(e, exc_info=True)
            self.logger.error('Error conection postgres DB ', self.conn)

        self.logger.info('done call get_conection')

    def get_subcategory_info_by_mappingCode(
            self, clase_predict: int = None) -> None:

        self.logger.info('call get_subcategory_info_by_mappingCode method ')
        self.get_conection()
        obj = None  

        try :
            with self.conn.cursor() as cur:

                postgres_select_query = '''

                            SELECT uid, name, keywords, category_uid, icon_uid
                            FROM public.subcategory
                            WHERE mapping_code=%(clase_predict)s;
                            '''

                param_select = {'clase_predict': clase_predict}
                cur.execute(postgres_select_query, param_select)

                obj = cur.fetchone()

        except Exception as e:
            self.logger.exception(e, exc_info=True)

        finally:
            self.conn.close()

            if obj is None:
                    self.logger.info(fmt('No found code subcategory: ',
                                    json={'code : ': clase_predict}))
                    return 0

            else :
                self.logger.info('done call get_subcategory_info_by_mappingCode')
                return obj

    def get_subcategory_info_batch_by_mappingCode(self, codes_predict) -> None:

        self.logger.info(
            'call get_subcategory_info_batch_by_mappingCode method ')
        self.get_conection()
        obj = None  

        try : 
            with self.conn.cursor() as cur:
                postgres_select_query = '''
                            SELECT uid, name, keywords, category_uid, icon_uid, mapping_code
                            FROM public.subcategory
                            WHERE mapping_code = ANY(%s);
                            '''
                cur.execute(postgres_select_query, (codes_predict,))
                # cur.execute('SELECT * FROM subcategory WHERE mapping_code = ANY (ARRAY[62,64])')
                # SELECT * FROM subcategory WHERE mapping_code = ANY (ARRAY[0,62,64])
                obj = cur.fetchall()

        except Exception as e:
            self.logger.exception(e, exc_info=True)

        finally:
            self.conn.close()
                    
            if obj is None:
                self.logger.info(fmt('No found code subcategory: ',
                                json={'code : ': codes_predict}))
                return 0

            else :
                self.logger.info(' done call get_subcategory_info_batch_by_mappingCode')
                return obj

    def get_category_info_by_uid(self, category_uid: str = None) -> None:

        self.logger.info('call get_category_info_by_uid method ')

        self.get_conection()
        obj = None  

        try :
            with self.conn.cursor() as cur:
                postgres_select_query = '''
                        SELECT name, keywords, color_uid, icon_uid, mapping_code
                        FROM public.category
                        WHERE uid=%(cat_uid)s;
                                        '''

                param_select = {'category_uid': category_uid}
                cur.execute(postgres_select_query, param_select)

                obj = cur.fetchone()

        except Exception as e:
            self.logger.exception(e, exc_info=True)

        finally:
            self.conn.close()

            if obj is None:
                self.logger.warning
                (
                    fmt('No found category_uid (get_category_info_by_uid): ', json={'category_uid': category_uid})
                )
                return 0
            else :
                self.logger.info('done call get_category_info_by_uid method ')
                return obj

    def get_subcategory_info_by_uid(self, subcategory_uid: str = None) -> None:

        self.logger.info('call get_subcategory_info_by_uid method ')

        self.get_conection()
        obj = None
        try :
            with self.conn.cursor() as cur:
                postgres_select_query = '''
                            SELECT mapping_code,
                                    name,
                                    keywords,
                                    category_uid
                            FROM public.subcategory
                            WHERE uid=%(subcategory_uid)s;
                            '''

                param_select = {'subcategory_uid': subcategory_uid}
                cur.execute(postgres_select_query, param_select)

                obj = cur.fetchone()

        except Exception as e:
            self.logger.exception(e, exc_info=True)
        finally:
            self.conn.close()
            # reveal_type(obj) would return 'Optional[Person]' here

            if obj is None:
                self.logger.warning(fmt('No found subcategory_uid(get_subcategory_info_by_uid): ',json={'subcategory_uid': subcategory_uid})
                )
                return 0
            else :
                self.logger.info('done call get_subcategory_info_by_uid method ')
                return obj

    def get_subcategory_info_by_parent(self, uid: str = None) -> None:

        self.logger.info('call get_subcategory_info_by_parent method ')

        self.get_conection()
        obj = None

        try :
            with self.conn.cursor() as cur:
                postgres_select_query = '''
                            SELECT uid,
                                    name
                            FROM public.subcategory
                            WHERE category_uid=%(uid)s;
                            '''

                param_select = {'uid': uid}
                cur.execute(postgres_select_query, param_select)

                obj = cur.fetchone()

        except Exception as e:
            self.logger.exception(e, exc_info=True)
        finally:
            self.conn.close()
            # reveal_type(obj) would return 'Optional[Person]' here

            if obj is None:
                self.logger.warning(fmt('No found subcategory_uid(get_subcategory_info_by_parent): ',json={'subcategory_uid': uid}))
                return 0
            else :
                self.logger.info('done call get_subcategory_info_by_parent method ')
                return obj

    def get_lang_subcategory_by_uid(self, subcategory_uid: str = None) -> None:

        self.logger.info('call get_lang_subcategory_by_uid method ')

        self.get_conection()
        obj = None  

        try :
            with self.conn.cursor() as cur:
                postgres_select_query = '''
                    SELECT uid, name, keywords, language_code, subcategory_uid
                    FROM public.subcategory_translation
                    WHERE subcategory_uid = %(subcatrgory_uid)s;
                                        '''
                param_select = {'subcategory_uid': subcategory_uid}
                cur.execute(postgres_select_query, param_select)

                obj = cur.fetchall()
        except Exception as e:
            self.logger.exception(e, exc_info=True)

        finally:
            self.conn.close()
            if obj is None:
                self.logger.warning(fmt('No found uid subcategory: ',json={'uid': subcategory_uid}))
                return 0
            else :
                self.logger.info('done call get_lang_subcategory_by_uid')
                return obj

    def get_bank_names(self, lang_code: str = 'es-MX') -> None:

        self.logger.info('call get_bank_names method')

        self.get_conection()
        obj = None  

        try :
            with self.conn.cursor() as cur:
                postgres_select_query = '''
                    SELECT name
                    FROM public.bank_name
                    WHERE lang_code = %(lang_code)s;
                                        '''

                param_select = {'lang_code': lang_code}

                cur.execute(postgres_select_query, param_select)

                obj = cur.fetchall()

        except Exception as e:
            self.logger.exception(e, exc_info=True)

        finally:
            self.conn.close()
            if obj is None:
                self.logger.info(fmt('No found get_bank_names for language code : ',json={'lang_code': lang_code}))
                return 0

            else :
                self.logger.info('done call get_bank_names')
                return obj

    def get_proper_name(self, lang_code: str = None) -> None:

        self.logger.info('call get_proper_name method')

        self.get_conection()
        obj = None  

        try :
            with self.conn.cursor() as cur:
                postgres_select_query = '''
                    SELECT name
                    FROM public.proper_name
                    WHERE lang_code = %(lang_code)s;
                                        '''

                param_select = {'lang_code': lang_code}
                cur.execute(postgres_select_query, param_select)
                # iterador ()
                obj = cur.fetchall()
        except Exception as e:
            self.logger.exception(e, exc_info=True)

        finally:
            self.conn.close()
            if obj is None:
                self.logger.info(
                    fmt(
                        'No found proper_name for language code : ',
                        json={'lang_code': lang_code},
                    )
                )
                return 0
            else :
                self.logger.info('done call proper_name')
                return obj

    def get_nicknames(self, country_code: str = None) -> None:

        self.logger.info('call get_nicknames method')

        self.get_conection()
        obj = None  

        try :
            with self.conn.cursor() as cur:
                postgres_select_query = '''
                    SELECT name
                    FROM public.nickname
                    WHERE country_code = %(country_code)s;
                                        '''
                
                param_select = {'country_code': country_code}
                cur.execute(postgres_select_query, param_select)

                obj = cur.fetchall()

        except Exception as e:
            self.logger.exception(e, exc_info=True)

        finally:
            self.conn.close()
            if obj is None:
                self.logger.info(
                    fmt(
                        'No found  nicknames for country code : ',
                        json={'country_code': country_code},
                    )
                )
                return 0
            else :
                self.logger.info('done call get_nicknames')
                return obj

    def get_months(self, lang_code: str = None) -> None:

        self.logger.info('call get_months method')

        self.get_conection()
        obj = None  

        try :
            with self.conn.cursor() as cur:
                postgres_select_query = '''
                    SELECT name
                    FROM public.month
                    WHERE lang_code = %(lang_code)s;
                                        '''
                param_select = {'lang_code': lang_code}

                cur.execute(postgres_select_query, param_select)

                obj = cur.fetchall()
        except Exception as e:
            self.logger.exception(e, exc_info=True)
        finally:
            self.conn.close()
            if obj is None:
                self.logger.info(
                    fmt(
                        'No found months for country code : ',
                        json={'country_code': lang_code},
                    )
                )
                return 0
            else :
                self.logger.info('done call get_months')
                return obj

    def get_places(self, lang_code: str = None) -> None:

        self.logger.info('call get_places method')

        self.get_conection()
        obj = None  

        try :
            with self.conn.cursor() as cur:
                postgres_select_query = '''
                    SELECT name
                    FROM public.place
                    WHERE lang_code = %(lang_code)s;
                                        '''
                param_select = {'lang_code': lang_code}

                cur.execute(postgres_select_query, param_select)

                obj = cur.fetchall()
        except Exception as e:
            self.logger.exception(e, exc_info=True)
        finally:
            self.conn.close()
            if obj is None:
                self.logger.info(
                    fmt(
                        'No found places for country code : ',
                        json={'country_code': lang_code},
                    )
                )
                return 0
            else :
                self.logger.info('done call get_places')
                return obj

    def get_keywords(self, country_code: str = None) -> None:

        self.logger.info('call get_keywords method')

        self.get_conection()
        obj = None  

        try :
            with self.conn.cursor() as cur:
                postgres_select_query = '''
                    SELECT name
                    FROM public.keyword
                    WHERE country_code = %(country_code)s;
                                        '''
                param_select = {'country_code': country_code}

                cur.execute(postgres_select_query, param_select)

                obj = cur.fetchall()

        except Exception as e:
            self.logger.exception(e, exc_info=True)

        finally:
            self.conn.close()
            if obj is None:
                self.logger.info(
                    fmt(
                        'No found keywords for country code : ',
                        json={'country_code': country_code},
                    )
                )
                return 0
            else:
                self.logger.info('done call get_keywords')
                return obj

    def get_stopword_context(self, lang_code: str = None) -> None:

        self.logger.info('call get_stopword_context method')

        self.get_conection()
        obj = None  

        try: 
            with self.conn.cursor() as cur:
                postgres_select_query = '''
                    SELECT name
                    FROM public.sw_context
                    WHERE lang_code = %(lang_code)s;
                                        '''
                param_select = {'lang_code': lang_code}

                cur.execute(postgres_select_query, param_select)

                obj = cur.fetchall()
        except Exception as e:
            self.logger.exception(e, exc_info=True)

        finally:
            self.conn.close()
            if obj is None:
                self.logger.info(
                    fmt(
                        'No found stopwords on context banking for language code : ',
                        json={'lang_code': lang_code},
                    )
                )
                return 0
            else:
                self.logger.info('done call get_stopword_context')
                return obj

    def get_catalog_merchant(self, country_code: str = None) -> None:

        self.logger.info('call get_catalog_merchant method')

        self.get_conection()
        obj = None  

        try:
            with self.conn.cursor() as cur:
                postgres_select_query = '''
                    SELECT nickname, collective_nickname, merchant_name,description
                    FROM public.merchant
                    WHERE country_code = %(country_code)s;
                                        '''
                param_select = {'country_code': country_code}
                cur.execute(postgres_select_query, param_select)
                obj = cur.fetchall()
        except Exception as e:
            self.logger.exception(e, exc_info=True)
        finally:
            self.conn.close()            
            if obj is None:
                self.logger.warning(
                    fmt(
                        'No found catalog merchant for country code : ',
                        json={'country_code': country_code},
                    )
                )
                return 0
            else :
                self.logger.info('done call get_catalog_merchant')
                return obj

    def get_info_merchant_by_nickname(self, nickname, country_code):

        self.logger.info('call get_merchant_name method')

        self.get_conection()
        obj = None  
        try:
            with self.conn.cursor() as cur:
                postgres_select_query = '''
                    SELECT collective_nickname,
                            merchant_name,
                            description,uid
                    FROM public.merchant
                    WHERE nickname = %(nickname)s AND
                        country_code = %(country_code)s;
                                        '''
                param_select = {
                    'nickname': nickname,
                    'country_code': country_code}

                cur.execute(postgres_select_query, param_select)

                obj = cur.fetchall()
        except Exception as e:
            self.logger.exception(e, exc_info=True)
        finally:
            self.conn.close()            
            if obj is None:
                self.logger.warning(
                    fmt(
                        'No found catalogs of nicknames and country code : ',
                        json={'country_code': country_code,
                              'nickname': nickname},
                    )
                )
                return 0
            else :
                self.logger.info('done call get_merchant_name')
                return obj

    def get_info_bacth_merchant_by_nickname(self, nicknames, country_code):

        self.logger.info('call get_info_bacth_merchant_by_nickname method')

        self.get_conection()
        obj = None  

        try :
            with self.conn.cursor() as cur:
                postgres_select_query = '''
                    SELECT nickname,
                            collective_nickname,
                            merchant_name,
                            description,
                            uid
                    FROM public.merchant
                    WHERE nickname = ANY (%s) AND
                        country_code = (%s);
                                        '''

                cur.execute(postgres_select_query, [nicknames, country_code])

                obj = cur.fetchall()
        except Exception as e:
            self.logger.exception(e, exc_info=True)

        finally:
            self.conn.close()  
            if obj is None:
                self.logger.info(
                    fmt(
                        'No found catalogs of nicknames and country code : ',
                        json={'country_code': country_code,
                              'nickname': nicknames},
                    )
                )
                return 0
            else:
                self.logger.info('done call get_info_bacth_merchant_by_nickname')
                return obj

    def get_info_merchant_by_uid(self, merch_uid, country_code):

        self.logger.info('call get_merchant_name method')

        self.get_conection()
        obj = None  

        try :
            with self.conn.cursor() as cur:
                postgres_select_query = '''
                    SELECT  nickname,
                            collective_nickname,
                            merchant_name,
                            description
                    FROM public.merchant
                    WHERE uid = %(merch_uid)s AND
                        country_code = %(country_code)s;
                                        '''
                param_select = {
                    'merch_uid': merch_uid,
                    'country_code': country_code}

                cur.execute(postgres_select_query, param_select)

                obj = cur.fetchall()
        except Exception as e:
            self.logger.exception(e, exc_info=True)
        finally:
            self.conn.close()  
            if obj is None:
                self.logger.info(
                    fmt(
                        'No found merchant : ',
                        json={'country_code': country_code,
                              'merch_uid': merch_uid},
                    )
                )
                return 0
            else :
                self.logger.info('done call get_merchant_name')
                return obj

    def get_info_batch_merchant_by_uids(self, merch_uids=None, country_code='MX'):

        self.logger.info('call get_merchant_name method')

        self.get_conection()
        obj = None  

        try :
            with self.conn.cursor() as cur:
                postgres_select_query = '''
                    SELECT  uid,
                            nickname,
                            collective_nickname,
                            merchant_name,
                            description
                    FROM public.merchant
                    WHERE country_code = (%s)
                    AND uid = ANY (%s) ;
                                        '''

                cur.execute(postgres_select_query, [country_code, merch_uids])

                obj = cur.fetchall()
        except Exception as e:
            self.logger.exception(e, exc_info=True)
                
        finally:
            self.conn.close()  
            if obj is None:
                self.logger.info(
                    fmt(
                        'No found merchant : ',
                        json={'country_code': country_code,
                              'merch_uid': merch_uids},
                    )
                )
                return 0
            else :
                self.logger.info('done call get_info_merchant_by_uids')
                return obj

    def get_custom_category_by_userUID(self, user_uid):

        self.logger.info('call get_custom_category_by_userUID method')

        self.get_conection()
        obj = None  

        try :
            with self.conn.cursor() as cur:
                postgres_select_query = '''
                    SELECT  uid,
                            name,
                            category_uid,
                            keywords

                    FROM public.custom_category
                    WHERE user_uid = %(user_uid)s
                                        '''
                param_select = {'user_uid': user_uid}

                cur.execute(postgres_select_query, param_select)

                obj = cur.fetchall()
        except Exception as e:
            self.logger.exception(e, exc_info=True)
        finally:
            self.conn.close()  
            if obj is None:
                self.logger.warning(
                    fmt('No found user_uid : ',
                        json={'user_uid': user_uid})
                )
                return 0
            else :
                self.logger.info('done call get_custom_category_by_userUID')
                return obj

    def insert_fact_table(self, list_items: list = None):
        self.logger.info('call insert_fact_table method')
        self.get_conection()
        try:
            with self.conn.cursor() as cur:
                for data in list_items:
                    postgres_insert_query = '''
                        INSERT INTO fact (
                            uid,
                            transaction_uid,
                            hash,
                            category_uid,
                            subcategory_uid,
                            custom_subcategory_uid,
                            lang_code,
                            country_code,
                            created_at,
                            updated_at,
                            valid_from,
                            valid_to,
                            user_uid,
                            description,
                            merchant_name,
                            stem,
                            bow,
                            collective_nickname,
                            nickname,
                            custom_nickname,
                            status
                            )
                        VALUES (
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s,
                                %s
                                )
                            '''
                    record_to_insert = (
                        generate(size=40),
                        data['transaction_uid'],
                        data['hash'],
                        data['category_uid'],
                        data['subcategory_uid'],
                        data['custom_subcategory_uid'],
                        data['lang_code'],
                        data['country_code'],
                        data['created_at'],
                        data['updated_at'],
                        data['valid_from'],
                        data['valid_to'],
                        data['user_uid'],
                        data['description'],
                        data['merchant_name'],
                        data['stem'],
                        data['bow'],
                        data['collective_nickname'],
                        data['nickname'],
                        data['custom_nickname'],
                        data['status'],
                    )
                    cur.execute(postgres_insert_query, record_to_insert)

        except BaseException as e:
            self.logger.exception(e, exc_info=True)
            self.logger.info(fmt('No insert transaction ', json={'data': list_items}))
            self.conn.rollback()
            return 0
        else:
            self.conn.commit()
        finally:
            self.conn.close()
            self.logger.info('done call insert_fact_table')
            return postgres_insert_query

    def insert_feature_store_table(self, list_items: list = None):
        '''
        '''
        self.logger.info('call insert_feature_store_table method')
        self.get_conection()
        try:
            with self.conn.cursor() as cur:
                for data in list_items:
                    postgres_insert_query = '''
                        INSERT INTO public.\
                                        feature_store
                                        (
                                            uid,
                                            description,
                                            hash,
                                            nickname,
                                            merchant_name,
                                            stem,
                                            bow,
                                            amount,
                                            category_uid,
                                            subcategory_uid,
                                            lang_code,
                                            country_code,
                                            single_cat,
                                            raw
                                        )
                            VALUES (%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s);
                                        '''
                    record_to_insert = (
                        generate(size=40),
                        data['description'],
                        data['hash'],
                        data['nickname'],
                        data['merchant_name'],
                        data['stem'],
                        data['bow'],
                        data['amount'],
                        data['category_uid'],
                        data['subcategory_uid'],
                        data['lang_code'],
                        data['country_code'],
                        data['single_cat'],
                        data['raw'],
                    )
                    cur.execute(postgres_insert_query, record_to_insert)
            
        except BaseException as e:
            self.logger.exception(e, exc_info=True)
            self.logger.info(fmt('No insert features ', json={'data': list_items}))
            self.conn.rollback()
            return 0
        else:
            self.conn.commit()
        finally:
            self.conn.close()
            self.logger.info('done call insert_feature_store_table')
            return postgres_insert_query           

    def insert_custom_category(self, data: dict = None):
        '''
        '''
        self.logger.info('call insert_custom_category method')
        self.get_conection()
        try:
            with self.conn.cursor() as cur:
                postgres_insert_query = '''
                INSERT INTO public.custom_category(
                uid, 
                name, 
                keywords, 
                user_uid, 
                created_at, 
                updated_at, 
                deleted_at)
                VALUES (%s,%s,%s,%s,%s,%s,%s)
                ON CONFLICT (uid) DO NOTHING
                                        '''

                record_to_insert = (
                    data['uid'],
                    data['name'],
                    data['keywords'],
                    data['user_uid'],
                    data['created_at'],
                    data['updated_at'],
                    data['deleted_at'])
                cur.execute(postgres_insert_query, record_to_insert)

        except BaseException as e:
            self.logger.exception(e, exc_info=True)
            self.logger.info(fmt('No insert new custom category ', json={'data': data}))
            self.conn.rollback()
            return 0
        else:
            self.conn.commit()
            return postgres_insert_query

        finally:
            self.conn.close()
            self.logger.info('done call insert_custom_category')
            
    def insert_custom_subcategory(self, data: dict = None):
        '''
        '''
        self.logger.info('call insert_custom_category method')
        self.get_conection()
        try:
            with self.conn.cursor() as cur:

                postgres_insert_query = '''
                INSERT INTO public.custom_subcategory(
                uid,
                category_uid,
                name,  
                keywords, 
                user_uid, 
                created_at, 
                updated_at, 
                deleted_at)
                VALUES (%s,%s,%s,%s,%s,%s,%s,%s)
                ON CONFLICT (uid) DO NOTHING
                                        '''

                record_to_insert = (
                    data['uid'],
                    data['parent_uid'],
                    data['name'],
                    data['keywords'],
                    data['user_uid'],
                    data['created_at'],
                    data['updated_at'],
                    data['deleted_at'])
                cur.execute(postgres_insert_query, record_to_insert)

        except BaseException as e:
            self.logger.exception(e, exc_info=True)
            self.logger.info(fmt('No insert new subcategory ', json={'data': data}))
            self.conn.rollback()
            return 0
        else:
            self.conn.commit()
            return postgres_insert_query

        finally:
            self.conn.close()
            self.logger.info('done call insert_custom_subcategory')
            
    def insert_custom_nickname(self, list_items: list = None):
        '''
        '''
        self.logger.info('call insert_custom_nickname method')
        self.get_conection()
        try:
            with self.conn.cursor() as cur:
                for data in list_items:

                    postgres_insert_query = '''
                    INSERT INTO public.custom_nickname(
                    uid, 
                    name, 
                    merchant_uid,
                    cuntry_code, 
                    keywords, 
                    user_uid, 
                    created_at, 
                    updated_at, 
                    deleted_at)
                    VALUES (%s,%s,%s,%s,%s,%s,%s,%s,%s)
                                            '''

                    record_to_insert = (
                        data['uid'],
                        data['name'],
                        data['merchant_uid'],
                        data['cuntry_code'],
                        data['keywords'],
                        data['user_uid'],
                        data['created_at'],
                        data['updated_at'],
                        data['deleted_at'])
                    cur.execute(postgres_insert_query, record_to_insert)

        except BaseException as e:
            self.logger.exception(e, exc_info=True)
            self.logger.info(fmt('No insert features ', json={'data': list_items}))
            self.conn.rollback()
            return 0
        else:
            self.conn.commit()
            return postgres_insert_query

        finally:
            self.conn.close()
            self.logger.info('done call insert_custom_nickname')

    def update_custom_category(self, data: dict = None):
        '''
        '''
        dt = datetime.now(timezone.utc).isoformat()

        data['created_at'] = dt
        data['updated_at'] = dt
        data['deleted_at'] = ''
        
        self.logger.info('call update_custom_category method')
        self.get_conection()
        try:
            with self.conn.cursor() as cur:
                postgres_insert_query = '''
                UPDATE public.custom_category SET 
                name=(%s),
                keywords=(%s),
                user_uid=(%s),
                created_at=(%s),
                updated_at=(%s),
                deleted_at=(%s)

                WHERE uid=(%s)
                
                        '''

                record_to_insert = (
                    data['name'],
                    data['keywords'],
                    data['user_uid'],
                    data['created_at'],
                    data['updated_at'],
                    data['deleted_at'],
                    data['uid'])
                cur.execute(postgres_insert_query, record_to_insert)

        except BaseException as e:
            self.logger.exception(e, exc_info=True)
            self.logger.info(fmt('No update custom category ', json={'data': data}))
            self.conn.rollback()
            return 0
        else:
            self.conn.commit()
            return postgres_insert_query

        finally:
            self.conn.close()
            self.logger.info('done call updated_custom_category')
    
    def update_custom_subcategory(self, data: dict = None):
        '''
        '''
        self.logger.info('call update_custom_subcategory method')
        self.get_conection()
        try:
            with self.conn.cursor() as cur:
                postgres_insert_query = '''
                UPDATE public.custom_subcategory SET 
                category_uid=(%s),
                name=(%s),
                icon_uid=(%s),
                color_uid=(%s),
                keywords=(%s),
                user_uid=(%s),
                created_at=(%s),
                updated_at=(%s),
                deleted_at=(%s)

                WHERE uid=(%s)
                
                '''
                
                record_to_insert = (
                    data['category_uid'],
                    data['name'],
                    data['icon_uid'],
                    data['color_uid'],
                    data['keywords'],
                    data['user_uid'],
                    data['created_at'],
                    data['updated_at'],
                    data['deleted_at'],
                    data['uid'])

                cur.execute(postgres_insert_query, record_to_insert)

        except BaseException as e:
            self.logger.exception(e, exc_info=True)
            self.logger.info(fmt('No update custom subcategory ', json={'data': data}))
            self.conn.rollback()
            return 0
        else:
            self.conn.commit()
            return postgres_insert_query

        finally:
            self.conn.close()
            self.logger.info('done call updated_custom_subcategory')

    def delete_custom_subcategory(self, custom_subcategory_uid: str = None) -> None:

        self.logger.info('call delete_custom_subcategory method ')

        self.get_conection()
        try :
            with self.conn.cursor() as cur:
                postgres_select_query = '''
                                        Delete from public.custom_subcategory where uid = (%s)
                                        '''

                record_to_insert = (custom_subcategory_uid,)
                cur.execute(postgres_select_query, record_to_insert)

        except Exception as e:
            self.logger.exception(e, exc_info=True)
            self.logger.info(
                    fmt('No found uid for custom subcategory: ',
                        json={'uid': custom_subcategory_uid})
                )
        else :
            self.conn.commit()
            self.logger.info('done call delete_custom_subcategory method ')

        finally:
            self.conn.close()
            
    def delete_custom_category(self, custom_category_uid: str = None) -> None:

        self.logger.info('call delete_custom_category method ')

        self.get_conection()
        try :
            with self.conn.cursor() as cur:
                postgres_select_query = '''
                                        Delete from public.custom_category where uid = (%s)
                                        '''
                record_to_insert = (custom_category_uid,)
                cur.execute(postgres_select_query, record_to_insert)

        except Exception as e:
            self.logger.exception(e, exc_info=True)
            self.logger.info(
                    fmt('No found uid for custom subcategory: ',
                        json={'uid': custom_category_uid})
                )
        else :
            self.conn.commit()
            self.logger.info('done call delete_custom_category method ')

        finally:
            self.conn.close()
     
    def get_custom_subcategory_info_by_uid(self, custom_subcategory_uid: str = None) -> None:

        self.logger.info('call get_custom_subcategory_uid_info_by_uid method ')

        self.get_conection()
        obj = None  

        try :
            with self.conn.cursor() as cur:
                postgres_select_query = '''

                            SELECT name,
                                    category_uid
                            FROM public.custom_subcategory
                            WHERE uid=%(custom_subcategory_uid)s;
                            '''

                param_select = {'custom_subcategory_uid': custom_subcategory_uid}
                cur.execute(postgres_select_query, param_select)

                obj = cur.fetchone()

        except Exception as e:
            self.logger.exception(e, exc_info=True)
        finally:
            self.conn.close()
            # reveal_type(obj) would return 'Optional[Person]' here

            if obj is None:
                self.logger.warning(
                    fmt('No found uid for custom subcategory: ',
                        json={'custom_subcategory_uid': custom_subcategory_uid})
                )
                return 0

            else :
                self.logger.info('done call get_custom_subcategory_uid_info_by_uid method ')
                return obj

    def get_custom_subcategory_info_by_parent(self, uid: str = None) -> None:

        self.logger.info('call get_custom_subcategory_info_by_parent method ')

        self.get_conection()
        obj = None  

        try :
            with self.conn.cursor() as cur:
                postgres_select_query = '''

                            SELECT uid,
                                    name
                            FROM public.custom_subcategory
                            WHERE category_uid=%(uid)s;
                            '''

                param_select = {'category_uid': uid}
                cur.execute(postgres_select_query, param_select)

                obj = cur.fetchone()

        except Exception as e:
            self.logger.exception(e, exc_info=True)
        finally:
            self.conn.close()
            # reveal_type(obj) would return 'Optional[Person]' here

            if obj is None:
                self.logger.warning(
                    fmt('No found custom_category_uid(get_custom_subcategory_info_by_parent): ',
                        json={'custom_category_uid': uid})
                )
                return 0

            else :
                self.logger.info('done call get_custom_subcategory_info_by_parent method ')
                return obj

    def what_is(self,uid):
        '''
        Fix Refill comm
        '''
        flag_subcategory = None
        subcategory_name = None
        subcategory_uid = None
        category_uid = None
        category_custom_type = None
        subcategory_custom_type = None
        obj = None
        
        # [mapping_code,name,keywords,category_uid]
        obj = self.get_subcategory_info_by_uid(uid)

        if obj != 0 and obj != None:
            flag_subcategory = True
            subcategory_uid = uid
            category_uid = obj[3]
            subcategory_name = obj[1]
            category_custom_type = 'core'
            subcategory_custom_type = 'core'
            self.logger.info('Found core subcategory ')

        
        elif obj == 0:
            # ['name','parant_iud']
            obj = self.get_custom_subcategory_info_by_uid(uid)
            
            if obj != 0 :
                flag_subcategory = True
                category_uid = obj[1]
                subcategory_uid = uid
                subcategory_name = obj[0]
                category_custom_type = 'custom'
                subcategory_custom_type = 'custom'
                self.logger.info('Found custom subcategory')

            elif  obj == 0:
                # response :[(subcategory_uid, name_subcategory)]
                self.get_subcategory_info_by_parent(uid)
                if obj != 0 :
                    flag_subcategory = False
                    category_uid = uid
                    subcategory_uid = obj[0]
                    subcategory_name = obj[1]
                    category_custom_type = 'core'
                    subcategory_custom_type = 'core'

                    self.logger.info('Found core category by parent')

            else:
                # response :[(subcategory_uid, name_subcategory)]
                self.get_custom_subcategory_info_by_parent(uid)
                if obj != 0 :
                    flag_subcategory = False
                    category_uid = uid
                    subcategory_uid = obj[0]
                    subcategory_name = obj[1]
                    subcategory_custom_type = 'custom'
                    category_custom_type = 'custom'

                    self.logger.info('Found custom subcategory by parent')


        return [category_uid,subcategory_uid,subcategory_name,category_custom_type,subcategory_custom_type,flag_subcategory]
    
    def find_category_uid(self,subcategory_uid):
        type_ = None
        obj = None
        
        obj = self.get_subcategory_info_by_uid(subcategory_uid)
        if obj != 0:
            uid_category = obj[3]
        type_ = 'core'
        
        
        if obj == 0:
            obj = self.get_custom_subcategory_info_by_uid(subcategory_uid)
            type_ = 'custom'

            if obj != 0:
                uid_category = obj[2] 
        
        return [uid_category,type_]
        
    def get_custom_subcat_name_by_uid(self, uid: str = None) -> None:

        self.logger.info("call get_subcategory_name_by_uid method ")

        self.get_conection()
        obj = None  

        with self.conn.cursor() as cur:
            postgres_select_query = """

                        SELECT name
                        FROM public.custom_category
                        WHERE uid=%(uid)s;
                        """

            param_select = {"uid": uid}
            cur.execute(postgres_select_query, param_select)

            obj = cur.fetchone()

            # reveal_type(obj) would return 'Optional[Person]' here

            if obj is None:
                self.logger.warning(
                    fmt("No found custom subcategory_uid (get_custom_subcat_name_by_uid): ",
                        json={"custom_subcategory_uid": uid})
                )
                return 0

            self.logger.info(
                "done call get_subcategory_name_by_uid method ")

            return obj
            
    def get_categorys_names(self) -> None:

        self.logger.info('call get_categorys_name method ')

        self.get_conection()
        obj = None  

        try :
            with self.conn.cursor() as cur:
                postgres_select_query = '''
                        SELECT uid, name
                        FROM public.category
                                        '''
                cur.execute(postgres_select_query)

                obj = cur.fetchall()

        except Exception as e:
            self.logger.exception(e, exc_info=True)

        finally:
            self.conn.close()

            if obj is None:
                self.logger.warning('No table category')
                return 0
            else :
                self.logger.info('done call get_categorys_name method ')
                return obj

    def get_subcategorys_names(self) -> None:

        self.logger.info('call get_subcategorys_names method ')

        self.get_conection()
        obj = None  

        try :
            with self.conn.cursor() as cur:
                postgres_select_query = '''
                            SELECT uid,
                                    name
                            FROM public.subcategory
                            '''

                cur.execute(postgres_select_query)

                obj = cur.fetchall()

        except Exception as e:
            self.logger.exception(e, exc_info=True)
        finally:
            self.conn.close()
            if obj is None:
                self.logger.info('No table subcategory')
                return 0

            else :
                self.logger.info('done call get_subcategorys_names method ')
                return obj



if __name__ == '__main__':
    pass
    # Controller = ControllerPosgrestDB()
    # Controller.get_conection()



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from pandas.tseries.offsets import MonthEnd
import matplotlib.cm as cm
import scipy.stats



def DetectRecurrencyI(
                      trans_data : pd.DataFrame,
                      amount_tolerance : float=0.01, 
                      period_tolerance : int=6,
                      n_days : int=3,
                      client_col: str =None,
                      time_col: str = None, 
                      amount_col: str = None,
                      config: dict = None
                      ):
    """It takes as arguments the dataframe of choosen a customer, an integer "period_tolerance" that is the minimum number of occuracy of an amount that it be selected, a float number "amount_tolerance" that is the accepted range for the amount, an integer "n_days" that is the accepted variance in the payement day. It returns a plot scatter of all the transactions where the noise amount (not recurring) are in grey and the  recurring amounts in different color."""
    if config is not None:
        client_col = config.customer_id
        amount_col = config.trans_amount
        time_col = config.trans_date

    # Creation of the new columns
    # number_occuracy_big, trans_data
    number_occuracy_big, account_table = _complete_table(trans_data, 
                                period_tolerance=period_tolerance,
                                client_col=client_col, 
                                time_col=time_col, 
                                amount_col=amount_col,
                                config=config)

    # Check the time rules for recurrency
    number_occuracy_big['day_of_month_cond'] = number_occuracy_big[amount_col].map(lambda x: 
                        (number_occuracy_big[(number_occuracy_big[amount_col] > x - x * amount_tolerance)
                                    & 
                    (number_occuracy_big[amount_col] <= x + x * amount_tolerance)]
                     .day_of_month.value_counts().count() < n_days))

    number_occuracy_big['number_day_end_month_cond'] = number_occuracy_big[amount_col].map(lambda x:
           (number_occuracy_big[(number_occuracy_big[amount_col] > x - x * amount_tolerance)
                                    & 
                    (number_occuracy_big[amount_col] <= x + x * amount_tolerance)]
                     .number_day_end_month.value_counts().count() < n_days))

    number_occuracy_big['number_business_day_end_month_cond'] = number_occuracy_big[amount_col].map(lambda x:
                  (number_occuracy_big[(number_occuracy_big[amount_col] >= x - x * amount_tolerance)
                                    & 
                    (number_occuracy_big[amount_col] <= x + x * amount_tolerance)]
                     .number_business_day_end_month.value_counts().count() < n_days))
    
    number_occuracy_big['rec'] = ((number_occuracy_big['day_of_month_cond'] == True) |
                        (number_occuracy_big['number_day_end_month_cond'] == True) |
                        (number_occuracy_big['number_business_day_end_month_cond'] == True))
        
    _plot_rec(account_table, 
            number_occuracy_big, 
            amount_tolerance =amount_tolerance ,
            period_tolerance=period_tolerance,
            client_col=client_col, 
            time_col=time_col, 
            amount_col=amount_col,
            config=config)




    return number_occuracy_big[number_occuracy_big['rec'] == True]





def DetectRecurrencyII(
                        trans_data : pd.DataFrame,
                        amount_tolerance : float=0.01,
                        period_tolerance : int=6, 
                        n_days : int=3,
                        client_col: str =None,
                        time_col: str = None, 
                        amount_col: str = None,
                        type_col:str = None,
                        config: dict = None
                        ):
    """It takes as arguments the dataframe of choosen customer, an integer "period_tolerance" that is the minimum number of occuracy of an amount that it be selected, a float number "amount_tolerance" that is the accepted range for the amount, an integer "n_days" that is the accepted variance in the payement day. It returns a list of couple (recurring amount, number of occuracy of the corresponding amount), and a dictionnary where the keys are the listed recurring amounts and it values are the corresponding dataframe."""
    
    if config is not None:
        client_col = config.customer_id
        amount_col = config.trans_amount
        time_col = config.trans_date

                                
    
    if isinstance(trans_data, pd.DataFrame):

        rec_amount_freq, rec_table_dict, rec_table_unnamed_dict, rec_amount_unnamed_freq = _detect_recurrency(trans_data,
                                                            amount_tolerance=amount_tolerance,
                                                            period_tolerance=period_tolerance,
                                                            client_col=client_col, 
                                                            time_col=time_col, 
                                                            amount_col=amount_col,
                                                            type_col=type_col,
                                                            config=config)

    else:

        rec_amount_freq = []
        rec_table_dict = []
        rec_table_unnamed_dict = []
        rec_amount_unnamed_freq = []

        for i in trans_data:
            # rec_table_unnamed_dict, rec_amount_unnamed_freq
            _rec_amount_freq, _rec_table_dict, _, _ = _detect_recurrency(i,
                                                            amount_tolerance=amount_tolerance,
                                                            period_tolerance=period_tolerance,
                                                            client_col=client_col, 
                                                            time_col=time_col, 
                                                            amount_col=amount_col,
                                                            type_col=type_col,
                                                            config=config)

            rec_amount_freq.append(_rec_amount_freq)
            rec_table_dict.append(_rec_table_dict)

    return  rec_amount_freq, rec_table_dict, rec_table_unnamed_dict, rec_amount_unnamed_freq





def _complete_table(
                    trans_data : pd.DataFrame,
                    amount_tolerance : float=0.01,
                    period_tolerance : int=5,
                    client_col: str =None,
                    time_col: str = None, 
                    amount_col: str = None,
                    config: dict = None
                   ):
    """
    It takes as arguments the dataframe of a choosen customer, an integer "period_tolerance" that is the minimum number of occuracy of an amount that it be selected, a float number "amount_tolerance" that is the accepted range for the amount, an integer "n_days" that is the accepted variance in the payement day. It returns the same table with additionnal columns that are needed by the functions "DetectRecurrencyI" and "DetectRecurrencyII".
    """    
    if config is not None:
        client_col = config.customer_id
        amount_col = config.trans_amount
        time_col = config.trans_date
        
    
    # Creation of the number of occuracy column
    # obtener resultados que están cerca del primer monto encontrado ()
    trans_data['freq'] = trans_data[amount_col].map(lambda x: 
                                        np.sum((trans_data[amount_col] >= x - x * amount_tolerance)
                                                        & 
                                                (trans_data[amount_col] <= x + x * amount_tolerance)))

    
    # Creation of the time columns
    trans_data[time_col] = trans_data[time_col].astype('datetime64')

    
    # El parámetro 0 en MonthEnd sólo especifica avanzar hasta el final del mes dado.
    # Tenga en cuenta que si hubiéramos usado MonthEnd(1), 
    # tendríamos la siguiente fecha, que es al final del mes.
    
    # último día del mes de la fecha en cuestion
    trans_data['end_of_month'] = pd.to_datetime(
        trans_data[time_col], format="%Y%m") + MonthEnd(0)
    
    # extraer el día 
    trans_data['day_of_month'] = trans_data[time_col].dt.day
    
    # número de días para finalizar el mes
    trans_data['number_day_end_month'] = trans_data.end_of_month.dt.day - trans_data.day_of_month
    
    # extraer sólo la fecha
    A = [d.date() for d in trans_data[time_col]]
    B = [d.date() for d in trans_data['end_of_month']]
    
    # cuenta el número de días válidos (días hábiles) dado el rango de fechas
    trans_data['number_business_day_end_month'] = np.busday_count(A, B)
    
    # Selecting the condidate for Recurring values 
    number_occuracy_big = trans_data[trans_data['freq'] >= period_tolerance].copy()

    return number_occuracy_big, trans_data



    
def _detect_recurrency(
                        trans_data : pd.DataFrame,
                        amount_tolerance : float=0.01,
                        period_tolerance : int=6, 
                        n_days : int=3,
                        client_col: str =None,
                        time_col: str = None, 
                        amount_col: str = None,
                        type_col:str = None,
                        config: dict = None
                        ):
        
    """It is a private function used by "DetectRecurrencyII". The function "_detect_recurrency" takes one dataframe and returns a list and a dictionary. The list contained the couple recurring amount and their frequency in the dataframe."""

    if config is not None:
        client_col = config.customer_id
        amount_col = config.trans_amount
        time_col = config.trans_date
        type_col = config.trans_type
  
    #  number_occuracy_big, trans_data
    trans_data, account_table = _complete_table(trans_data, 
                                amount_tolerance =amount_tolerance ,
                                period_tolerance=period_tolerance,
                                client_col=client_col, 
                                time_col=time_col, 
                                amount_col=amount_col,
                                config=config)
    

    trans_data = trans_data.set_index([amount_col]).sort_index()
    
    

    X_unique = set(trans_data.index)

    rec_amount_freq = []
    rec_table_dict = {}
    
    rec_amount_unnamed_freq = []
    rec_table_unnamed_dict = {}
    # loop over the amounts
    for x in X_unique:

        rec = None

        # Table of the wanted amount related x
        # A = trans_data.loc[pd.Interval(x - amount_tolerance * x, x + amount_tolerance * x)]
        A = trans_data.loc[x - x *amount_tolerance:  x + x * amount_tolerance]

        # FOR CREDIT TRANSACTION
        if ((A.shape[0] >= period_tolerance) & (A[type_col].value_counts().count() == 1)):

            # most freq day happens at least 3 times
            day_of_month_cond = A.day_of_month.value_counts().count() < n_days
            
            number_day_end_month_cond = A.number_day_end_month.value_counts().count() < n_days
            
            number_business_day_end_month_cond = A.number_business_day_end_month.value_counts().count() < n_days
            
            nickname = A.nickname.unique().tolist()
            
            A.sort_values(by='value_date',ascending=False,inplace=True)
            
            #A['days_diff'] = A.groupby(['user_uid','nickname'])['value_date'].diff().apply(lambda x: x/np.timedelta64(-1, 'D')).fillna(0).astype('int64')
            
            
            # This table will be selected if one of the previous condition is true
            if (day_of_month_cond | number_day_end_month_cond | number_business_day_end_month_cond):
                rec = A
            
            list_day_diff = A.days_diff.values
            mean_day_diff = scipy.stats.mode(list(filter(lambda x: x > n_days, list_day_diff)))[0][0]
            
            if rec is not None:
                if len(nickname) == 1:
                    rec_amount_freq.append((nickname[0], x, A.shape[0], mean_day_diff))
                    rec_table_dict[(nickname[0], x, A.shape[0])] = rec
                else :
                    rec_amount_unnamed_freq.append(('_'.join(nickname), x, A.shape[0], mean_day_diff))
                    rec_table_unnamed_dict[('_'.join(nickname), x, A.shape[0])] = rec

    return rec_amount_freq, rec_table_dict, rec_table_unnamed_dict, rec_amount_unnamed_freq




def _plot_rec(
            account_table : pd.DataFrame,
            number_occuracy_big : pd.DataFrame,
            amount_tolerance : float=0.01,
            period_tolerance : int = 6,
            client_col: str =None,
            time_col: str = None, 
            amount_col: str = None,
            config: dict = None
            ):

    """it takes as argument a panda series (a column), bins and ranges and plot a histogram."""
    
    
    if config is not None:
        client_col = config.customer_id
        amount_col = config.trans_amount
        time_col = config.trans_date
    
    
  #  number_occuracy_big = account_table[account_table['rec'] == True]
    plt.style.use('seaborn-deep')
    fig, ax = plt.subplots(figsize=(20, 10))


    x2=account_table[account_table['freq'] < period_tolerance][time_col]
    y2=account_table[account_table['freq'] < period_tolerance][amount_col]

    rec = set(number_occuracy_big[number_occuracy_big['rec'] == True][amount_col])
    colors = cm.rainbow(np.linspace(0, 1, len(rec)))

    for  x, col in zip(rec, colors):
        x1=number_occuracy_big[(number_occuracy_big[amount_col] >= x - x * amount_tolerance)
                                                            & 
                            (number_occuracy_big[amount_col] <= x + x * amount_tolerance)][time_col]
        y1=number_occuracy_big[(number_occuracy_big[amount_col] >= x - x * amount_tolerance)
                                                            & 
                            (number_occuracy_big[amount_col] <= x + x * amount_tolerance)][amount_col]

        ax.plot(x1, y1, 's', color=col, markersize=5)
    ax.plot(x2, y2, 'o', markerfacecolor='black', markersize=6,  alpha=0.5)

    ax.set_xlabel('Date of transaction', fontsize=20)
    ax.set_ylabel('Amount of transaction', fontsize=20)

    ax.grid()

    plt.xticks(fontsize=14, rotation=30)
    plt.yticks(fontsize=14, rotation=0)
    plt.show()
    return 



